@@ -39,4 +39,7 @@ int md_major(void);
 int blkext_major(void);
 int max_partitions(int major);
 
+int dev_subsystem_part_major(const struct device *dev);
+const char *dev_subsystem_name(const struct device *dev);
+
 #endif
