@@ -42,11 +42,16 @@ extern char *optarg;
 #  define OPTIND_INIT 1
 #endif
 
+#ifdef UDEV_SYNC_SUPPORT
+#  define LIBUDEV_I_KNOW_THE_API_IS_SUBJECT_TO_CHANGE
+#  include <libudev.h>
+#endif
+
 /*
  * Table of valid switches
  */
 static struct arg _the_args[ARG_COUNT + 1] = {
-#define arg(a, b, c, d, e) {b, "", "--" c, d, e, 0, NULL, 0, 0, INT64_C(0), UINT64_C(0), SIGN_NONE, PERCENT_NONE, NULL},
+#define arg(a, b, c, d, e) {b, "", "--" c, d, e, 0, NULL, 0, 0, INT64_C(0), UINT64_C(0), SIGN_NONE, PERCENT_NONE},
 #include "args.h"
 #undef arg
 };
@@ -54,61 +59,60 @@ static struct arg _the_args[ARG_COUNT + 1] = {
 static struct cmdline_context _cmdline;
 
 /* Command line args */
-/* FIXME: struct cmd_context * is unnecessary (large # files ) */
+/* FIXME: Move static _the_args into cmd? */
 unsigned arg_count(const struct cmd_context *cmd __attribute((unused)), int a)
 {
 	return _the_args[a].count;
 }
 
+unsigned arg_is_set(const struct cmd_context *cmd, int a)
+{
+	return arg_count(cmd, a) ? 1 : 0;
+}
+
 const char *arg_value(struct cmd_context *cmd __attribute((unused)), int a)
 {
 	return _the_args[a].value;
 }
 
-const char *arg_str_value(struct cmd_context *cmd __attribute((unused)),
-			  int a, const char *def)
+const char *arg_str_value(struct cmd_context *cmd, int a, const char *def)
 {
 	return arg_count(cmd, a) ? _the_args[a].value : def;
 }
 
-int32_t arg_int_value(struct cmd_context *cmd __attribute((unused)),
-		      int a, const int32_t def)
+int32_t arg_int_value(struct cmd_context *cmd, int a, const int32_t def)
 {
 	return arg_count(cmd, a) ? _the_args[a].i_value : def;
 }
 
-uint32_t arg_uint_value(struct cmd_context *cmd __attribute((unused)),
-			int a, const uint32_t def)
+uint32_t arg_uint_value(struct cmd_context *cmd, int a, const uint32_t def)
 {
 	return arg_count(cmd, a) ? _the_args[a].ui_value : def;
 }
 
-int64_t arg_int64_value(struct cmd_context *cmd __attribute((unused)),
-			int a, const int64_t def)
+int64_t arg_int64_value(struct cmd_context *cmd, int a, const int64_t def)
 {
 	return arg_count(cmd, a) ? _the_args[a].i64_value : def;
 }
 
-uint64_t arg_uint64_value(struct cmd_context *cmd __attribute((unused)),
-			  int a, const uint64_t def)
+uint64_t arg_uint64_value(struct cmd_context *cmd, int a, const uint64_t def)
 {
 	return arg_count(cmd, a) ? _the_args[a].ui64_value : def;
 }
 
-const void *arg_ptr_value(struct cmd_context *cmd __attribute((unused)),
-			  int a, const void *def)
+/* No longer used.
+const void *arg_ptr_value(struct cmd_context *cmd, int a, const void *def)
 {
 	return arg_count(cmd, a) ? _the_args[a].ptr : def;
 }
+*/
 
-sign_t arg_sign_value(struct cmd_context *cmd __attribute((unused)),
-		      int a, const sign_t def)
+sign_t arg_sign_value(struct cmd_context *cmd, int a, const sign_t def)
 {
 	return arg_count(cmd, a) ? _the_args[a].sign : def;
 }
 
-percent_t arg_percent_value(struct cmd_context *cmd __attribute((unused)),
-			    int a, const percent_t def)
+percent_t arg_percent_value(struct cmd_context *cmd, int a, const percent_t def)
 {
 	return arg_count(cmd, a) ? _the_args[a].percent : def;
 }
@@ -180,21 +184,7 @@ int yes_no_excl_arg(struct cmd_context *cmd __attribute((unused)),
 
 int metadatatype_arg(struct cmd_context *cmd, struct arg *a)
 {
-	struct format_type *fmt;
-	char *format;
-
-	format = a->value;
-
-	dm_list_iterate_items(fmt, &cmd->formats) {
-		if (!strcasecmp(fmt->name, format) ||
-		    !strcasecmp(fmt->name + 3, format) ||
-		    (fmt->alias && !strcasecmp(fmt->alias, format))) {
-			a->ptr = fmt;
-			return 1;
-		}
-	}
-
-	return 0;
+	return get_format_by_name(cmd, a->value) ? 1 : 0;
 }
 
 static int _get_int_arg(struct arg *a, char **ptr)
@@ -361,6 +351,9 @@ int int_arg_with_sign_and_percent(struct cmd_context *cmd __attribute((unused)),
 	else if (!strcasecmp(ptr, "F") || !strcasecmp(ptr, "FR") ||
 		 !strcasecmp(ptr, "FREE"))
 		a->percent = PERCENT_FREE;
+	else if (!strcasecmp(ptr, "O") || !strcasecmp(ptr, "OR") ||
+		 !strcasecmp(ptr, "ORIGIN"))
+		a->percent = PERCENT_ORIGIN;
 	else
 		return 0;
 
@@ -453,10 +446,7 @@ int alloc_arg(struct cmd_context *cmd __attribute((unused)), struct arg *a)
 
 int segtype_arg(struct cmd_context *cmd, struct arg *a)
 {
-	if (!(a->ptr = (void *) get_segtype_from_string(cmd, a->value)))
-		return 0;
-
-	return 1;
+	return get_segtype_from_string(cmd, a->value) ? 1 : 0;
 }
 
 /*
@@ -803,7 +793,7 @@ static int _get_settings(struct cmd_context *cmd)
 			  "be activated read-only.");
 	}
 
-	if (arg_count(cmd, ignorelockingfailure_ARG))
+	if (arg_count(cmd, ignorelockingfailure_ARG) || arg_count(cmd, sysinit_ARG))
 		init_ignorelockingfailure(1);
 	else
 		init_ignorelockingfailure(0);
@@ -907,11 +897,53 @@ static void _apply_settings(struct cmd_context *cmd)
 
 	set_activation(cmd->current_settings.activation);
 
-	cmd->fmt = arg_ptr_value(cmd, metadatatype_ARG,
-				 cmd->current_settings.fmt);
+	cmd->fmt = get_format_by_name(cmd, arg_str_value(cmd, metadatatype_ARG,
+				      cmd->current_settings.fmt_name));
+
 	cmd->handles_missing_pvs = 0;
 }
 
+static int _set_udev_checking(struct cmd_context *cmd)
+{
+#ifdef UDEV_SYNC_SUPPORT
+	struct udev *udev;
+	const char *udev_dev_dir;
+	size_t udev_dev_dir_len;
+	int dirs_diff;
+
+	if (!(udev = udev_new()) ||
+	    !(udev_dev_dir = udev_get_dev_path(udev)) ||
+	    !*udev_dev_dir) {
+		log_error("Could not get udev dev path.");
+		return 0;
+	}
+	udev_dev_dir_len = strlen(udev_dev_dir);
+
+	/* There's always a slash at the end of dev_dir. But check udev_dev_dir! */
+	if (udev_dev_dir[udev_dev_dir_len - 1] != '/')
+		dirs_diff = strncmp(cmd->dev_dir, udev_dev_dir,
+				    udev_dev_dir_len);
+	else
+		dirs_diff = strcmp(cmd->dev_dir, udev_dev_dir);
+
+	if (dirs_diff) {
+		log_debug("The path %s used for creating device nodes and "
+			  "symlinks that is set in the configuration differs "
+			  "from the path %s that is used by udev. All warnings "
+			  "about udev not working correctly while processing "
+			  "particular nodes and symlinks will be suppressed. "
+			  "These nodes and symlinks will be managed in each "
+			  "directory separately.",
+			   cmd->dev_dir, udev_dev_dir);
+		dm_udev_set_checking(0);
+		init_udev_checking(0);
+	}
+
+	udev_unref(udev);
+#endif
+	return 1;
+}
+
 static const char *_copy_command_line(struct cmd_context *cmd, int argc, char **argv)
 {
 	int i, space;
@@ -1006,6 +1038,9 @@ int lvm_run_command(struct cmd_context *cmd, int argc, char **argv)
 	log_debug("O_DIRECT will be used");
 #endif
 
+	if (!_set_udev_checking(cmd))
+		goto_out;
+
 	if ((ret = _process_common_commands(cmd)))
 		goto_out;
 
@@ -1014,9 +1049,7 @@ int lvm_run_command(struct cmd_context *cmd, int argc, char **argv)
 	else
 		locking_type = -1;
 
-	if (!init_locking(locking_type, cmd)) {
-		log_error("Locking type %d initialisation failed.",
-			  locking_type);
+	if (!init_locking(locking_type, cmd, arg_count(cmd, sysinit_ARG))) {
 		ret = ECMD_FAILED;
 		goto out;
 	}
@@ -1054,10 +1087,16 @@ int lvm_run_command(struct cmd_context *cmd, int argc, char **argv)
 	dm_pool_empty(cmd->mem);
 
 	reset_lvm_errno(1);
+	reset_log_duplicated();
 
 	return ret;
 }
 
+int lvm_return_code(int ret)
+{
+	return (ret == ECMD_PROCESSED ? 0 : ret);
+}
+
 int lvm_split(char *str, int *argc, char **argv, int max)
 {
 	char *b = str, *e;
@@ -1196,6 +1235,10 @@ static void _fin_commands(void)
 		dm_free(_cmdline.commands[i].valid_args);
 
 	dm_free(_cmdline.commands);
+
+	_cmdline.commands = NULL;
+	_cmdline.num_commands = 0;
+	_cmdline.commands_size = 0;
 }
 
 void lvm_fin(struct cmd_context *cmd)
@@ -1246,7 +1289,7 @@ static int _run_script(struct cmd_context *cmd, int argc, char **argv)
 		ret = lvm_run_command(cmd, argc, argv);
 		if (ret != ECMD_PROCESSED) {
 			if (!error_message_produced()) {
-				log_debug("Internal error: Failed command did not use log_error");
+				log_debug(INTERNAL_ERROR "Failed command did not use log_error");
 				log_error("Command failed with status code %d.", ret);
 			}
 			break;
@@ -1322,6 +1365,10 @@ int lvm2_main(int argc, char **argv)
 		unsetenv("LVM_DID_EXEC");
 	}
 
+	/* "version" command is simple enough so it doesn't need any complex init */
+	if (!alias && argc > 1 && !strcmp(argv[1], "version"))
+		return lvm_return_code(version(NULL, argc, argv));
+
 	if (!(cmd = init_lvm()))
 		return -1;
 
@@ -1371,13 +1418,11 @@ int lvm2_main(int argc, char **argv)
 		log_error("No such command.  Try 'help'.");
 
 	if ((ret != ECMD_PROCESSED) && !error_message_produced()) {
-		log_debug("Internal error: Failed command did not use log_error");
+		log_debug(INTERNAL_ERROR "Failed command did not use log_error");
 		log_error("Command failed with status code %d.", ret);
 	}
 
       out:
 	lvm_fin(cmd);
-	if (ret == ECMD_PROCESSED)
-		ret = 0;
-	return ret;
+	return lvm_return_code(ret);
 }
