@@ -94,7 +94,6 @@ static void _pvscan_display_single(struct cmd_context *cmd,
 		  display_size(cmd, (uint64_t) (pv_pe_count(pv) -
 						pv_pe_alloc_count(pv)) *
 					   pv_pe_size(pv)));
-	return;
 }
 
 int pvscan(struct cmd_context *cmd, int argc __attribute((unused)),
@@ -153,7 +152,7 @@ int pvscan(struct cmd_context *cmd, int argc __attribute((unused)),
 		/* Also check for MD use? */
 /*******
 		if (MAJOR(pv_create_kdev_t(pv[p]->pv_name)) != MD_MAJOR) {
-			log_print
+			log_warn
 			    ("WARNING: physical volume \"%s\" belongs to a meta device",
 			     pv[p]->pv_name);
 		}
@@ -167,7 +166,7 @@ int pvscan(struct cmd_context *cmd, int argc __attribute((unused)),
 			size_new += pv_size(pv);
 			size_total += pv_size(pv);
 		} else
-			size_total += pv_pe_count(pv) * pv_pe_size(pv);
+			size_total += (uint64_t) pv_pe_count(pv) * pv_pe_size(pv);
 	}
 
 	/* find maximum pv name length */
